/********************************************************************************************
 *  Ebbie-MM Version 4.2
 *  Copyright (c)  2008 April  H. Alexander Ebhardt, Herbert H. Tsang and Denny Chen Dai
 *  Permission is granted to copy, distribute and/or modify this document
 *  under the terms of the GNU Free Documentation License, Version 1.2
 *  or any later version published by the Free Software Foundation;
 *  with no Invariant Sections, no Front-Cover Texts, and no Back-Cover
 *  Texts.  A copy of the license is included in the section entitled "GNU
 *  Free Documentation License".
 *********************************************************************************************/

package ebbieMM.align;

import ebbieMM.util.RString;

import ebbieMM.util.ResultLog;

import java.util.Vector;



/*

 * Alignment.java

 * Created by Denny Chen Dai 

 * A Java Class Representation of Alignments objects 

 */

public class Alignment {

   

    private long queryHead = -1; // start position of the aligned segment in query sequence

    private long queryTail = -1; // end pos of the aligned segment in query sequence 

    public long sbjctHead = -1; //start pos of the segment in db sequence

    public long sbjctTail = -1;// end pos of the segment in db sequence 

    public int mis_pos = -1; //the pos of the  mismatch under the condition of one mismatch

                              //otherwise -1

    public Vector qbase = new Vector(); //query alignment string 

    public Vector sbase = new Vector(); //db sequence alignment string    

    public Vector mutLog = new Vector(); //a mutation log in the format " a -> b .." that has performed

                                           //on this class 

    public float score; //the alignment score in blast output 

    

     

    /** Creates a new instance of Alignment */

    public Alignment() {

    }

    

    //initialize the alignment with a max size array 

    public Alignment(long size){       

    }

    

    public String getQbase(){

        String tmp = "";

        for(int i=0; i< this.qbase.size();i++)

            tmp = tmp + qbase.get(i);

        return tmp;

    }

  

   /*

    * perform a one-time mutation at pos , log it

    */

    public String mutate(int pos){



        String old = (String)qbase.get(pos);

        this.mutLog.clear();

        

        String tmp = ((String)qbase.get(pos)).toUpperCase();      

        if(tmp.equals("A") )

            return this.mutateQuery(pos,"C");

        else

        if(tmp.equals("C" ))

            return this.mutateQuery(pos,"G");

        else

        if(tmp.equals("G"))

            return this.mutateQuery(pos,"T");

        else

        if(tmp.equals("T"))

            return this.mutateQuery(pos,"A");

        return "";

    }

    

    /*

     * mutate qbase at pos into 'residue' character

     */

    private String mutateQuery(int pos, String residue){



        String tmp = qbase.get(pos).toString().toUpperCase() + "->" + residue + "(" + pos  + ")";

        mutLog.add(tmp);

        qbase.set(pos,residue);   

        return tmp;

    }

    

    /**
     * We check whether the alignment itself is a full match 
     *@param ori_length The original lenght of the query sequence
     *@return The number of mismatch, 0 indicates a full match, 1 indicates a one-pos mismatch
     */

    public int checkFullMatch(int ori_length){

        

        if(qbase.size() < ori_length)  //obviously not 

              return -1;

          

        int misCount = 0; //number of mismatched  

        String tmp1, tmp2;

        for(int i=0; i<qbase.size();i++){

            tmp1 = ((String)qbase.get(i)).toUpperCase();

            tmp2 = ((String)sbase.get(i)).toUpperCase();

            if(!tmp1.equals(tmp2)){

                misCount++;

                this.mis_pos = i; 

           }

        }

        return misCount;

    }

    

    /**

     * XML version of parsing in blast output for one single alignment result

     * 

     * blocks within   <Hsp>  and </Hsp>

     * Example

     * 

     *  <Hsp>

              <Hsp_num>1</Hsp_num>

              <Hsp_bit-score>34.1929</Hsp_bit-score>

              <Hsp_score>17</Hsp_score>

              <Hsp_evalue>0.000702077</Hsp_evalue>

              <Hsp_query-from>1</Hsp_query-from>

              <Hsp_query-to>21</Hsp_query-to>

              <Hsp_hit-from>774906</Hsp_hit-from>

              <Hsp_hit-to>774926</Hsp_hit-to>

              <Hsp_query-frame>1</Hsp_query-frame>

              <Hsp_hit-frame>1</Hsp_hit-frame>

              <Hsp_identity>20</Hsp_identity>

              <Hsp_positive>20</Hsp_positive>

              <Hsp_align-len>21</Hsp_align-len>

              <Hsp_qseq>CCTTTTAATTATTTATTGGGA</Hsp_qseq>

              <Hsp_hseq>CCTTTTGATTATTTATTGGGA</Hsp_hseq>

              <Hsp_midline>|||||| ||||||||||||||</Hsp_midline>

            </Hsp>

     * 

     * @param data - xml chunck containg one alignment statistics

     */

    public void parseXML(String data){

        int start=0,end=0;

        /*obtain score*/

        start = data.indexOf("<Hsp_bit-score>", 0) + "<Hsp_bit-score>".length();

        end = data.indexOf("</Hsp_bit-score>", start);

        score = Float.parseFloat(data.substring(start, end));

        

        /*query sequence begin pos*/

        start = data.indexOf("<Hsp_query-from>", 0) + "<Hsp_query-from>".length();

        end = data.indexOf("</Hsp_query-from>", start);

        queryHead = Long.parseLong(data.substring(start, end));

        

        /*query sequence end pos*/

        start = data.indexOf("<Hsp_query-to>", 0) + "<Hsp_query-to>".length();

        end = data.indexOf("</Hsp_query-to>", start);

        queryTail = Long.parseLong(data.substring(start, end));

        

        /*query string itself*/

        start = data.indexOf("<Hsp_qseq>", 0) + "<Hsp_qseq>".length();

        end = data.indexOf("</Hsp_qseq>", start);

        String tmp = data.substring(start, end);

        for(int i=0;i<tmp.length();i++)

            qbase.add(tmp.substring(i,i+1));

        

        /*db hit begin pos*/

        start = data.indexOf("<Hsp_hit-from>", 0) + "<Hsp_hit-from>".length();

        end = data.indexOf("</Hsp_hit-from>", start);

        sbjctHead = Long.parseLong(data.substring(start, end));

        

        /*db hit end pos*/

        start = data.indexOf("<Hsp_hit-to>", 0) + "<Hsp_hit-to>".length();

        end = data.indexOf("</Hsp_hit-to>", start);

        sbjctTail = Long.parseLong(data.substring(start, end));

        

        /*db string itself*/

        start = data.indexOf("<Hsp_hseq>", 0) + "<Hsp_hseq>".length();

        end = data.indexOf("</Hsp_hseq>", start);

        tmp = data.substring(start, end);

        for(int i=0;i<tmp.length();i++)

            sbase.add(tmp.substring(i,i+1));

    }

    

/*

 * Parse in the raw blast output string and populate one alignment data 

 *

 * Score = 22.3 bits (11), Expect = 3.0

 * Identities = 11/11 (100%)

 * Strand = Plus / Minus

 * Query: 4      caaagagtaga 14

 *             |||||||||||

 * Sbjct: 749769 caaagagtaga 749759

 * Score

*/ 

    public String parse(RString data){

        

        String tmp = ""; //parsing start right after "Score =  " "    

        score = Float.parseFloat(data.GetNextWord()); // the score value 

        //==now let's get to the "Query: " part

        do{

            tmp = data.GetNextWord();

        }while(!(tmp.equals("Query")));

        

        tmp = data.GetNextToken(); // ":"

        queryHead = Long.parseLong(data.GetNextWord());

        do{

            qbase.add(data.GetNextToken());

            tmp = data.SeeNextToken(); //read in query sequence

        }while(!(tmp.charAt(0) >= '0' && tmp.charAt(0) <='9'));

        queryTail = Long.parseLong(data.GetNextWord());

        

        //========================================================

        do{

            tmp = data.GetNextWord();

        }while(tmp.charAt(0) == '|');

        //=========================================================

        tmp = data.GetNextWord(); // : 

        sbjctHead = Long.parseLong(data.GetNextWord());

        do{

            sbase.add(data.GetNextToken());

            tmp = data.SeeNextToken();

        }while(!(tmp.charAt(0) >= '0' && tmp.charAt(0) <='9'));

        

        sbjctTail = Long.parseLong(data.GetNextWord());

        return data.GetNextWord();

    }

    

    /**

     * Display the alignment into log file 

     */

    public void displayAlignment(ResultLog out){ 

        String qtmp="",stmp= "";

        String matchString = "", queryString ="", dbString = "";

        for(int i=0; i< qbase.size();i++){

            qtmp = (String)qbase.get(i);

            stmp = (String)sbase.get(i);

            

            queryString += qtmp;

            dbString += stmp; 

            if(qtmp.equals(stmp))

                matchString += "|";

            else

                matchString += "&";

        }    

        out.println(queryString + 

                "\t ( " + this.queryHead + " -- " + this.queryTail + " ) " + 

                "\n" + 

                matchString + 

                "\n" + 

                dbString + 

                "\t ( " + this.sbjctHead + " -- " + this.sbjctTail + " ) " 

                );

        

    }

    

    /*

     * Display mutation history on the alignment 

     */

    public void displayMutation(ResultLog out){

        for(int i=0; i< mutLog.size();i++){        

            out.println("#" + (String)mutLog.get(i) );

        }

    }



}



