/********************************************************************************************
 *  Ebbie-MM Version 4.2
 *  Copyright (c)  2008 Nov.  H. Alexander Ebhardt, Herbert H. Tsang and Denny Chen Dai
 *  Permission is granted to copy, distribute and/or modify this document
 *  under the terms of the GNU Free Documentation License, Version 1.2
 *  or any later version published by the Free Software Foundation;
 *  with no Invariant Sections, no Front-Cover Texts, and no Back-Cover
 *  Texts.  A copy of the license is included in the section entitled "GNU
 *  Free Documentation License".
 *********************************************************************************************/

package ebbieMM;

import com.sun.corba.se.impl.orbutil.concurrent.Mutex;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Vector;

/**
 *
 * Spawning class for blastn execution 
 * 
 *
 */
public class BlastNSpawner {
    
 
    public final Vector eater_flag = new Vector();
    String db_file = ""; //the database file name with full directory
    
    /** Creates a new instance of RNADistanceSpawner */
    public BlastNSpawner(String db) {
        db_file = db;
    }
    
    /**
     * Spawn feeder & eater thread for BlastN
     * @param query_buffer The input query sequence buffer.
     * @param result_buffer The blast output buffer
     */
    public void spawnBlastN(final Vector query_buffer, final Vector result_buffer,
            final Mutex mtx, boolean setPenalty, float penalty){
                
        query_buffer.clear();
        result_buffer.clear();
                
        String os = System.getProperty("os.name");
        String cmd = "";
        ProcessBuilder pb = null;
        
        /*determine OS type*/
        if (os.startsWith("Windows"))
            cmd = "blast\\blastall.exe";
        else
            cmd = "blast/./blastall";
       
        if(setPenalty)
            pb = new ProcessBuilder( cmd,
                "-p", "blastn",
                "-d", db_file,
                "-m","7",
                "-q",  String.valueOf(penalty));
        else
            pb = new ProcessBuilder(cmd,
                "-p", "blastn",
                "-d", db_file,
                "-m","7");
            
        //pb.directory(new File("blast"));
        pb.redirectErrorStream( true );

        try
            {
            // trigger the spawn, creating Process object for controlling child.
            final Process p = pb.start();
            
            // Spawn thread to write input to the spawned program
            new Thread()
                    
            {
            public void run()
                {
                // hook into input to spawned program
                final OutputStream os = p.getOutputStream();
                final OutputStreamWriter osw = new OutputStreamWriter( os );
                final BufferedWriter bw = new BufferedWriter( osw, 4028/* buffsize in chars */ );
                
                try{                 
                    /*process buffer whenever it is not empty*/
                    boolean escape = false;
                    while(!escape){  
                    while(query_buffer.size()!=0){                    
                        try{
                            /*load in the query sequence, remove from buffer*/                        
                            mtx.acquire();
                            String tmp = (String)query_buffer.remove(0);
                            mtx.release();                         
                            /*check termination indicator*/
                            if(tmp.charAt(0) == '@')
                            {
                                escape = true;
                                eater_flag.add("TRUE");//one item indicate eater to escape!!
                                break;
                            }                               
                            /*pump out data to Blast*/
                            bw.write( tmp + "\n" + "\n" );                            
                            System.out.println("pump sequence" + tmp);                     
                            bw.flush();

                        }catch(InterruptedException e){
                            System.err.println("Mutexing error in Spawner..");
                        }
                    }//inner while
                    
                    }//outer while
                    
                    /*end of process for BlastNspawner, close write buffer, terminate process*/
                    bw.close();
                    System.out.println("BlastNSpawner feeder exiting...");
                }
                catch ( IOException e )
                {
                    System.err.println( "problem writing spawn input" + e.getMessage() );
                }
                
                }//run
            }.start();
            
                     
            /*****************************************/
            //Spawn thread to read output of spawned program
            new Thread()
            {
            @SuppressWarnings( { "EmptyCatchBlock" } )
            public void run()
                {
                // hook into output from spawned program
                final InputStream is = p.getInputStream();
                final InputStreamReader isr = new InputStreamReader( is );
                final BufferedReader br = new BufferedReader( isr, 4028/* buffsize in chars */ );
                String line ="",buffer="";
                boolean escape = false;
                try
                {        
                  while(!escape){    
                      
                      /*obtain one blast output trunk*/
                      if(br.ready()){                      
                          
                          buffer = "";
                          boolean trunk_end = false;
                          
                          do{
                              /*read in one line*/
                              if ( ( line = br.readLine() ) != null )
                                  buffer += line + "\n";                      
                              /*end of stream reached*/
                              if(line == null){
                                  System.out.println("end of pipe stream...");
                                  break;
                              }
                              /*check end of output file*/
                              else
                              if(line.indexOf("</BlastOutput>")!=-1)
                                  trunk_end = true;
                          }while(!trunk_end);
                          
                          //System.err.println("finish reading lines..");
                          //System.out.println(buffer);
                                                                       
                          /*add to result buffer*/
                          try{  
                              mtx.acquire();
                              result_buffer.add(buffer);
                              System.out.println("received blasted result...");
                              mtx.release();   
                          }catch(InterruptedException e){
                              System.err.println("Error mtx..");
                          }
                      }
                      
                      /*check the flag to quit...*/   
                      if(eater_flag.size() > 0){
                          escape = true;
                          break;
                      }
                  }
                  br.close();
                  System.out.println("BlastNSpawner eater exiting...");
                  eater_flag.remove(0);
                }
                catch ( IOException e )
                {
                    System.err.println( "problem reading spawn output" + e.getMessage() );
                }
            }//run
            }.start();


            }
        catch ( IOException e )
        {
            System.err.println( "spawning problem " + e.getMessage() );
        }
    }
}
